mod tests;

pub mod p_var {
    #![allow(non_snake_case)]
    use std::fmt::{self, Display, Formatter};

    #[derive(Debug, PartialEq)]
    pub enum PVarError {
        EmptyArray,
        PRange,
    }

    impl Display for PVarError {
        fn fmt(&self, f: &mut Formatter) -> fmt::Result {
            match self {
                Self::EmptyArray => write!(f, "input array is empty"),
                Self::PRange => write!(f, "exponent must be greater or equal than 1.0"),
            }
        }
    }

    pub fn p_var_backbone<'a>(l: usize, p: f64, dist: &'a [Vec<f64>]) -> Result<f64, PVarError> {
        if p < 1. {
            return Err(PVarError::PRange);
        }
        if l == 0 {
            return Err(PVarError::EmptyArray);
        }

        if l == 1 {
            return Ok(0.);
        }

        let mut run_pvar = vec![0f64; l];
        let mut N = 1;
        let s = l - 1;

        while (s >> N) > 0 {
            N += 1;
        }

        let mut radius = vec![0f64; s];
        let ind_n = |j, n| -> usize { (s >> n) + (j >> n) };
        let center = |j, n| -> usize { ((j >> n) << n) + (1usize << (n - 1)) };
        let center_outside_range = |j, n| (j >> n == s >> n && (s >> (n - 1)) % 2usize == 0usize);

        let mut point_links = vec![0usize; l];
        let mut max_p_var = 0f64;

        for j in 0..l {
            for n in 1..=N {
                if !center_outside_range(j, n) {
                    let r = &mut radius[ind_n(j, n)];
                    *r = f64::max(*r, dist[center(j, n)][j]);
                }
            }
            if j == 0 {
                continue;
            }

            let mut m = j - 1;
            point_links[j] = m;

            let mut delta = dist[m][j];

            max_p_var = run_pvar[m] + delta.powf(p);

            let mut n = 0;

            while m > 0 {
                while n < N && (m >> n) % 2 == 0 {
                    n += 1;
                }
                m -= 1;

                let mut delta_needs_update = true;
                while n > 0 {
                    if !center_outside_range(m, n) {
                        let id = radius[ind_n(m, n)] + dist[center(m, n)][j];
                        if delta >= id {
                            break;
                        } else if delta_needs_update {
                            delta = (max_p_var - run_pvar[m]).powf(1f64 / p);
                            delta_needs_update = false;
                            if delta >= id {
                                break;
                            }
                        }
                    }
                    n -= 1;
                }
                if n > 0 {
                    m = (m >> n) << n;
                } else {
                    let d = dist[m][j];
                    if d >= delta {
                        let new_p_var = run_pvar[m] + d.powf(p);
                        if new_p_var >= max_p_var {
                            max_p_var = new_p_var;
                            point_links[j] = m;
                        }
                    }
                }
            }

            run_pvar[j] = max_p_var;
        }
        Ok(max_p_var)
    }

    pub fn p_var_backbone_ref<'a>(
        l: usize,
        p: f64,
        dist: &'a [Vec<f64>],
    ) -> Result<f64, PVarError> {
        if l == 0 {
            return Err(PVarError::EmptyArray);
        }
        if l == 1 {
            return Ok(0.);
        }

        let mut cum_p_var = vec![0f64; l];

        for j in 1..l {
            for m in 0..j {
                cum_p_var[j] = f64::max(cum_p_var[j], cum_p_var[m] + dist[m][j].powf(p));
            }
        }

        Ok(*cum_p_var.last().unwrap())
    }
}

use pyo3::prelude::*;

#[pyfunction]
#[pyo3(name = "p_var")]
fn pvar_wrapper(l: usize, p: f64, dist: Vec<Vec<f64>>) -> f64 {
    p_var::p_var_backbone(l, p, &dist).unwrap()
}

#[pymodule]
#[pyo3(name = "p_var")]
fn pvar(_py: Python<'_>, m: &PyModule) -> PyResult<()> {
    m.add_function(wrap_pyfunction!(pvar_wrapper, m)?)?;
    Ok(())
}
